package Kinoteka; # $Id: Kinoteka.pm 229 2021-10-25 20:08:54Z minus $
use strict;
use warnings FATAL => 'all';
use utf8;

binmode STDERR, ":raw:utf8";

=encoding utf8

=head1 NAME

Kinoteka - Minus's Film Library

=head1 VERSION

Version 1.00

=head1 SYNOPSIS

    <IfModule mod_perl.c>
        #PerlSwitches -I/path/to/kinoteka/lib
        PerlModule Kinoteka
        LogLevel perl:debug
        PerlOptions +GlobalRequest
        PerlOptions +ParseHeaders

        <Location /kinoteka/api>
            PerlInitHandler Kinoteka
            PerlSetVar Location /kinoteka/api
            #PerlSetVar PackshotDir /var/lib/kinoteka
            #PerlSetVar Debug off

            # Authentication & Authorization
            #AuthType Basic
            #AuthName "Kinoteka restricted zone"
            #AuthUserFile "/var/lib/kinoteka/.htpasswd"
            #<RequireAny>
            #    Require valid-user
            #    Require ip 127.0.0.1
            #    Require ip ::1
            #    #Require ip 192.168.0.0/16
            #</RequireAny>
            #
            # Note!! Run following command for add admin user
            #     htpasswd -Bb /var/lib/kinoteka/.htpasswd admin admin
            # ... and custom users:
            #     htpasswd -B /var/lib/kinoteka/.htpasswd username
            #
        </Location>
    </IfModule>

    # Static shared files
    Alias /kinoteka /usr/share/kinoteka
    <Directory /usr/share/kinoteka>
        Options Indexes FollowSymLinks
        AllowOverride None
        Require all granted
    </Directory>
    <Location /kinoteka/packshots>
        RedirectMatch "\/kinoteka\/packshots\/(.*)$" "/packshots/$1"
    </Location>
    Alias /packshots /var/lib/kinoteka/packshots
    <Directory /var/lib/kinoteka/packshots>
        Options Indexes FollowSymLinks
        AllowOverride None
        Require all granted
    </Directory>

    curl -v http://localhost/kinoteka

=head1 DESCRIPTION

Kinoteka is Minus's Film Library

=head2 SQLITE DDL

    CREATE TABLE "kinoteka" (
      "id"          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
      "obsolete"    INTEGER DEFAULT NULL,
      "storage"     INTEGER DEFAULT NULL,
      "name"        CHAR(255) DEFAULT NULL,
      "nameorig"    CHAR(255) DEFAULT NULL,
      "genre"       CHAR(255) DEFAULT NULL,
      "date"        INTEGER DEFAULT NULL,
      "imdbid"      CHAR(20) DEFAULT NULL,
      "kinopoiskid" CHAR(20) DEFAULT NULL,
      "country"     CHAR(255) DEFAULT NULL,
      "director"    CHAR(255) DEFAULT NULL,
      "scriptwriter" CHAR(255) DEFAULT NULL,
      "producer"    CHAR(255) DEFAULT NULL,
      "operator"    CHAR(255) DEFAULT NULL,
      "music"       CHAR(255) DEFAULT NULL,
      "artdirector" CHAR(255) DEFAULT NULL,
      "cut"         CHAR(255) DEFAULT NULL,
      "time"        CHAR(255) DEFAULT NULL,
      "ratingimdb"  REAL DEFAULT NULL,
      "ratingkinopoisk" REAL DEFAULT NULL,
      "description" TEXT DEFAULT NULL,
      "packshot"    TEXT DEFAULT NULL
    );

=head2 MYSQL DDL

    CREATE DATABASE `kinoteka` /*!40100 DEFAULT CHARACTER SET utf8 COLLATE utf8_bin */;
    CREATE TABLE `kinoteka` (
      `id` int(11) NOT NULL AUTO_INCREMENT,
      `obsolete` int(11) DEFAULT NULL,
      `storage` int(11) DEFAULT NULL,
      `name` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `nameorig` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `genre` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `date` year(4) DEFAULT NULL,
      `imdbid` varchar(20) COLLATE utf8_bin DEFAULT NULL,
      `kinopoiskid` varchar(20) COLLATE utf8_bin DEFAULT NULL,
      `country` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `director` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `scriptwriter` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `producer` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `operator` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `music` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `artdirector` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `cut` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `time` varchar(255) COLLATE utf8_bin DEFAULT NULL,
      `ratingimdb` float DEFAULT NULL,
      `ratingkinopoisk` float DEFAULT NULL,
      `description` text COLLATE utf8_bin,
      `packshot` text COLLATE utf8_bin,
      PRIMARY KEY (`id`),
      UNIQUE KEY `id` (`id`),
      FULLTEXT KEY `search` (`name`,`nameorig`,`description`)
    ) ENGINE=MyISAM DEFAULT CHARSET=utf8 COLLATE=utf8_bin;

=head1 MPMINUS METHODS

List of available MPMinus methods

=head2 handler

See L<https://metacpan.org/pod/MPMinus::REST>

=head2 hInit

See L<https://metacpan.org/pod/MPMinus::REST>

=head2 hLog

See L<https://metacpan.org/pod/MPMinus::REST>

=head2 error_response

See L<https://metacpan.org/pod/MPMinus::REST>

=head1 APPLICATION METHODS

=head2 request_filter

    my $handler_status = $server->request_filter();

This method validates input data structure (deserialized json document) and returns
boolean status of this operation. Also this method prepares the request data structure

=head2 response_filter

    my $handler_status = $server->response_filter();

This method validates outout data structure (serialized json document) and returns
boolean status of this operation. Also this method prepares the response data structure

=head2 instance

    my $instance = $server->instance;

Returns current DB instance

=head1 CONFIGURATION

Configuration performs by editing the standard configuration files
of the Apache WEB server, including .htaccess files of your project.

=head2 COMMON DIRECTIVES

=over 8

=item APIkey

Defines key (token) for API service that provides getting (grabbing) data about films

    APIkey 524e6340-5c6b-428c-b4bc-77dd4ac54e45

Default: none

See also the C<_grab> internal function

=item Debug

    PerlSetVar Debug on

Turns on debug mode

Default: off

=item Location

Defines a unique tag-name that marks the current location (Location,
Directory or File configuration block) of the handler, for example:

    <Location ~ ^/(kinoteka|filmoteka|$)>
        PerlInitHandler Kinoteka
        PerlSetVar Location /kinoteka/api
        ...
    </Location>

Default: none

B<Note!> Please do not forget specify this directive!

=item PackshotDir

Defines path to packshot directory

    PackshotDir /var/lib/kinoteka/foo

Default: /var/lib/kinoteka

=back

=head2 DB INSTANCE DIRECTIVES

=over 8

=item Kinoteka_Instance

    PerlSetVar Kinoteka_Instance foo

Sets the DB instance name

Default: none

=item Kinoteka_DB_DSN

    PerlSetVar Kinoteka_DB_DSN "DBI:mysql:database=kinoteka;host=mysql.example.com"

Sets the DBI DSN for database

More DBI examples:

    DBI:Pg:dbname=kinoteka;host=mysql.example.com
    DBI:SQLite:dbname=/var/tmp/mydb.db

Default: "DBI:Sponge:"

=item Kinoteka_DB_UserName, Kinoteka_DB_Password

    PerlSetVar Kinoteka_DB_UserName DBUsername
    PerlSetVar Kinoteka_DB_Password DBPassword

Sets the DB credentials

Default: none

=item Kinoteka_DB_TimeOut

    PerlSetVar Kinoteka_DB_TimeOut 30

Sets the timeout value for any database operations

Default: 60

=item Kinoteka_DB_Attributes

    PerlSetVar Kinoteka_DB_Attributes "RaiseError=0, PrintError=0"

Sets DB Attributes. This is string of pairs "name=value",
separated by commas char and space char(s)

Default: none

For MySQL recomends: "RaiseError=0, PrintError=0, mysql_enable_utf8=1"

=back

=head1 HISTORY

See C<Changes> file

=head1 DEPENDENCIES

L<MPMinus::REST|https://metacpan.org/pod/MPMinus::REST>

=head1 TO DO

See C<TODO> file

=head1 SEE ALSO

L<https://metacpan.org/pod/MPMinus::REST>

=head1 AUTHOR

Serż Minus (Sergey Lepenkov) L<https://www.serzik.com> E<lt>minus@serzik.comE<gt>

=head1 COPYRIGHT

Copyright (C) 1998-2021 D&D Corporation. All Rights Reserved

=head1 LICENSE

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

See C<LICENSE> file

=cut

use vars qw/ $VERSION /;
$VERSION = '1.00';

use Encode;
use Encode::Locale;

use base qw/MPMinus::REST/;

use Apache2::Util;
use Apache2::Const -compile => qw/ :common :http :satisfy /;
use Apache2::Access;
use Apache2::ServerUtil qw//;
use APR::Const qw//;
use APR::Table qw//;

use CGI qw//;
use URI;
use File::Spec;
use File::Copy qw//;
use LWP::UserAgent;
use List::Util qw/uniq/;

use CTK::Util qw/ sharedstatedir touch dtf trim dformat /;
use CTK::ConfGenUtil;
use CTK::TFVals qw/ :ALL /;
use CTK::Serializer;

use MPMinus::Util qw/ getHiTime getSID /;
use MPMinus::Store::DBI;

use constant {
        MP2                     => (exists $ENV{MOD_PERL_API_VERSION} && $ENV{MOD_PERL_API_VERSION} == 2) ? 1 : 0,
        IS_WIN32                => ($^O =~ /Cygwin|Win32/i ? 1 : 0),
        PREFIX                  => "kinoteka",
        CONTENT_TYPE            => "application/json",
        DATE_FORMAT             => "%Y-%m-%dT%H:%M:%SZ", # See strftime(3)
        METHOD_DEFAULT          => "default",
        DEFAULT_DSN_MASK        => 'DBI:SQLite:dbname=%s',
        DBFILE_MOD              => 0666,
        PACKSHOT_MAXSIZE        => 1024*1024, # 1MB limit
        PACKSHOT_DIR            => "packshots",
        XDEPTH                  => 3, # Packshot depth
        UA_TIMEOUT              => 20,
        ROWS_PER_PAGE           => 30,
        ROWS_PER_PAGE_MAX       => 1000,
        API_URL_GENERAL         => "https://kinopoiskapiunofficial.tech/api/v2.2/films/%d",

        # Serializer
        SR_ATTRS => {
                json => [
                        { # Out (serialize)
                            utf8 => 0,
                            pretty => 1,
                            allow_nonref => 1,
                            allow_blessed => 1,
                        },
                        { # In (deserialize)
                            utf8 => 0,
                            pretty => 1,
                            allow_nonref => 1,
                            allow_blessed => 1,
                        },
                    ],
            },

        # Directives of the dir-config (also from .htacces)
        DB_TIMEOUT => 60,
        DIRECTIVES => {
            debug                   => 'off',
            location                => 'default',
            packshotdir             => '',
            apikey                  => '',

            # DB
            kinoteka_instance       => "",
            kinoteka_db_dsn         => "", # "DBI:Sponge:",
            kinoteka_db_username    => "",
            kinoteka_db_password    => "",
            kinoteka_db_timeout     => 60,
            kinoteka_db_attributes  => "", # Recomended: "RaiseError => 0, PrintError => 0"
        },
    };

use constant DDL_CREATE_TABLE => <<'DDL';
CREATE TABLE "kinoteka" (
  "id"          INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE,
  "obsolete"    INTEGER DEFAULT NULL,
  "storage"     INTEGER DEFAULT NULL,
  "name"        CHAR(255) DEFAULT NULL,
  "nameorig"    CHAR(255) DEFAULT NULL,
  "genre"       CHAR(255) DEFAULT NULL,
  "date"        INTEGER DEFAULT NULL,
  "imdbid"      CHAR(20) DEFAULT NULL,
  "kinopoiskid" CHAR(20) DEFAULT NULL,
  "country"     CHAR(255) DEFAULT NULL,
  "director"    CHAR(255) DEFAULT NULL,
  "scriptwriter" CHAR(255) DEFAULT NULL,
  "producer"    CHAR(255) DEFAULT NULL,
  "operator"    CHAR(255) DEFAULT NULL,
  "music"       CHAR(255) DEFAULT NULL,
  "artdirector" CHAR(255) DEFAULT NULL,
  "cut"         CHAR(255) DEFAULT NULL,
  "time"        CHAR(255) DEFAULT NULL,
  "ratingimdb"  REAL DEFAULT NULL,
  "ratingkinopoisk" REAL DEFAULT NULL,
  "description" TEXT DEFAULT NULL,
  "packshot"    TEXT DEFAULT NULL
)
DDL

use constant DML_GET_COUNT => <<'DML';
SELECT COUNT(`id`) AS `cnt`
FROM `kinoteka`
DML

use constant DML_STORAGE_LAST => <<'DML';
SELECT MAX(`storage`) AS `storage_last`
FROM `kinoteka`
DML

use constant DML_GET_LAST_LIST => <<'DML';
SELECT *
FROM `kinoteka`
ORDER BY `id` DESC
LIMIT 56
DML

use constant DML_INSERT_EMPTY_RECORD => <<'DML';
INSERT INTO `kinoteka`
 (`storage`,`name`)
VALUES
 (?,?)
DML

use constant DML_SET_FILM => <<'DML';
UPDATE `kinoteka`
SET `obsolete` = ?, `storage` = ?, `name` = ?, `nameorig` = ?,
  `genre` = ?, `date` = ?, `imdbid` = ?, `kinopoiskid` = ?, `country` = ?,
  `director` = ?, `scriptwriter` = ?, `producer` = ?,
  `operator` = ?, `music` = ?, `artdirector` = ?, `cut` = ?,
  `time` = ?, `ratingimdb` = ?, `ratingkinopoisk` = ?, `description` = ?
WHERE `id` = ?
DML

use constant DML_SET_PACKSHOT => <<'DML';
UPDATE `kinoteka`
SET `packshot` = ?
WHERE `id` = ?
DML

use constant DML_SEARCH_COUNT => <<DML;
SELECT COUNT(`id`) AS `cnt`
FROM `kinoteka`
WHERE [WHERE]
DML

use constant DML_SEARCH_FILMS => <<DML;
SELECT *
FROM `kinoteka`
WHERE [WHERE]
ORDER BY `id` DESC
LIMIT [LIMIT]
DML

use constant DML_GENRES => <<DML;
SELECT DISTINCT LOWER(`genre`) FROM `kinoteka` WHERE `genre` IS NOT NULL ORDER BY LOWER(`genre`)
DML

use constant DML_COUNTRIES => <<DML;
SELECT DISTINCT `country` FROM `kinoteka` WHERE `country` IS NOT NULL ORDER BY `country`
DML

use constant DML_DELETE_FILM => <<DML;
DELETE
FROM `kinoteka`
WHERE id = ?
DML

my %instances; # List of DBI instance
my $xdb_errors = 0; # XDB initialize errors

sub handler : method {
    my $class = (@_ >= 2) ? shift : __PACKAGE__;
    my $r = shift;
    my $ct = $r->headers_in->get("content-type") || CONTENT_TYPE; # $r->headers_in->get("accept") ||
       $r->headers_in->set("content-type" => CONTENT_TYPE) if $ct =~ /json/i;

    return $class->init($r,
        sr_attrs => SR_ATTRS,
        dvars => DIRECTIVES,
    );
}
sub hInit {
    my $self = shift;
    my $r = shift;

    # Default systemdirs
    $self->set_dvar("datadir", File::Spec->catdir(sharedstatedir(), PREFIX));
    $self->set_dvar("packshotdir", File::Spec->catdir($self->get_dvar("datadir"), PACKSHOT_DIR))
        unless $self->get_dvar("packshotdir");

    # Dir config variables
    my $db_timeout = $self->get_dvar(PREFIX . "_db_timeout");
    $self->set_dvar(PREFIX . "_db_timeout", DB_TIMEOUT) unless is_int($db_timeout);

    # Set date string
    $self->set_svar(datef => decode(locale => Apache2::Util::ht_time($r->pool, time(), DATE_FORMAT, 1)));
    $self->set(to_datef => sub {
        decode(locale => Apache2::Util::ht_time($r->pool, shift || time(), DATE_FORMAT, 1));
    });

    # Make instance name and instance file
    my $instance_name = $self->get_dvar( PREFIX . "_instance" ); # $self->location
    my $instance_file = File::Spec->catfile($self->get_dvar("datadir"), sprintf("%s.db", $instance_name || PREFIX));
    $self->set_dvar("instance_file", $instance_file);

    # Make DSN string
    my $db_dsn = $self->get_dvar( PREFIX . "_db_dsn" ) || sprintf(DEFAULT_DSN_MASK, $instance_file);
    $self->set_dvar(
        instance_db => join("/", $r->server->server_hostname(), "db", $instance_name ? $instance_name : $db_dsn)
    );

    # DB Instance init
    my $instance_db_name = $self->get_dvar("instance_db");
    my $instance = $instances{$instance_db_name};
    my $do_init = 0;
    if(defined($instance)) {
        $instance->reconnect unless $instance->ping; # The connection is live
    } else {
        # Attributes prepare
        my %db_attr = _parseDBAttributes($self->get_dvar( PREFIX . "_db_attributes" ));
        $db_attr{RaiseError} //= 0;
        $db_attr{PrintError} //= 0;

        # SQLite
        if ($db_dsn =~ /SQLite/i) { # If SQLite (default)
            unless ((-e $instance_file) && !(-z $instance_file)) {
                touch($instance_file);
                chmod(DBFILE_MOD, $instance_file);
                $do_init = 1;
            }
            $db_attr{sqlite_unicode} //= 1;
        }

        # Make instance
        $instances{$instance_db_name} = $instance = MPMinus::Store::DBI->new(
            -dsn        => $db_dsn,
            -user       => $self->get_dvar( PREFIX . "_db_username" ),
            -pass       => $self->get_dvar( PREFIX . "_db_password" ),
            -connect_to => $self->get_dvar( PREFIX . "_db_timeout" ),
            -request_to => $self->get_dvar( PREFIX . "_db_timeout" ),
            -attr       => { %db_attr },
        );
        #print STDERR Data::Dumper::Dumper($instances{$instance_db_name});
    }

    # Check instance
    my $dbh = $instance->connect;
    if (!$dbh) {
        $self->error(Apache2::Const::SERVER_ERROR, $instance->error || $DBI::errstr || sprintf("Can't connect to %s", $db_dsn));
        delete $instances{$instance_db_name};
    } elsif ($do_init) {
        $instance->execute(DDL_CREATE_TABLE);
        if ($dbh->err) {
            $self->error(Apache2::Const::SERVER_ERROR, $dbh->errstr() || sprintf("Can't init %s", $db_dsn));
            delete $instances{$instance_db_name};
        }
    } elsif (!$instance->ping) {
        if ($instance->{dsn} !~ /Sponge/ig) {
            $self->error(Apache2::Const::HTTP_SERVICE_UNAVAILABLE, $instance->error || sprintf("Can't connect to %s", $db_dsn));
            delete $instances{$instance_db_name};
        }
    }

    # Origin
    my $rc = $self->SUPER::hInit($r);

    # SID renew
    $self->set_svar(sid => getSID(16, "m"));

    return $rc;
}
sub hLog {
    my $self = shift;
    my $r = shift;

    # Logging to system log
    if ($self->get_svar("debug")) {
        my $msg = sprintf(
            "[http %d] [uri=%s; handler=%s; sid=%s; status=%s; time=%.3f] %s",
            $self->code, $r->uri // '', $self->get_svar("name") // "", $self->get_svar("sid"),
            $self->status ? "OK" : "ERROR",
            sprintf("%.3f", (getHiTime() - $self->get_svar('hitime'))*1) || 0,
            $self->error // "-",
        );
        $self->error ? $r->log->error($msg) : $r->log->debug($msg);
    }

    #return $self->SUPER::hLog($r);
    return Apache2::Const::OK;
}

#
# MPMinus filters
#

sub request_filter {
    my $self = shift;
    my $req_data = shift || $self->get("req_data");

    # Correct output
    if (is_hash($req_data)) {
        $req_data->{"method"} //= METHOD_DEFAULT;
        $req_data->{"date"} //= $self->get_svar("datef");
    } else {
        return $self->error(Apache2::Const::HTTP_BAD_REQUEST => "Request invalid. Please check the request json document");
    }

    $self->set(req_data => $req_data);
    return 1; # Or 0 only!!
}
sub response_filter {
    my $self = shift;
    my $res_data = shift || $self->get("res_data");

    # Correct output
    if (is_hash($res_data)) {
        $res_data->{"id"} //= $self->get_svar("sid");
        $res_data->{"message"} //= "Ok";
        $res_data->{"date"} //= $self->get("to_datef")->(time());
        $res_data->{"time"} //= sprintf("%.3f", (getHiTime() - $self->get_svar('hitime')));
        $res_data->{"version"} //= $self->VERSION;
    } else {
        return $self->error(Apache2::Const::SERVER_ERROR => "Response invalid. Please check the response json document");
    }

    $self->set(res_data => $res_data);
    return 1; # Or 0 only!!
}
sub error_response {
    my $self = shift;
    return {
        "id"        => $self->get_svar("sid"),
        "message"   => $self->error || "unknown error",
        "date"      => $self->get("to_datef")->(time()),
        "time"      => sprintf("%.3f", (getHiTime() - $self->get_svar('hitime'))),
        "version"   => $self->VERSION,
    };
}
sub instance {
    my $self = shift;
    my $instance_db_name = $self->get_dvar("instance_db");
    return $instances{$instance_db_name};
}

__PACKAGE__->register_handler( # GET /
    handler => "index",
    method  => "GET",
    path    => "/",
    query   => undef,
    attrs   => {
            foo             => 'My foo attribute value',
            #debug           => 'on',
            content_type    => CONTENT_TYPE,
            #deserialize     => 1,
            serialize       => 1,
        },
    description => "Information data",
    code    => sub {
    my $self = shift;
    my $name = shift;
    my $r = shift;
    my $q = $self->get("q");
    my $usr = $self->get("usr");
    #my $req_data = $self->get("req_data");
    #my $res_data = $self->get("res_data");

    # Test database
    my $instance_db_name = $self->get_dvar("instance_db");
    my $instance = $instances{$instance_db_name};
    my $file = "";
    my $db_error;
    if (defined($instance) && $self->_is_sqlite) {
        $file = $instance->connect->sqlite_db_filename();
    } elsif (defined($instance)) { # Other
        $db_error = "Unsupported database driver";
    }

    # Get count
    my $cnt = 0;
    my $storage = 0;
    my @genres = ();
    my @countries = ();
    if ($instance) {
        # Get count
        $cnt = $instance->field(DML_GET_COUNT) || 0;
        $db_error = $instance->error if $instance->err;

        # Get storage last
        for ($instance->table(DML_STORAGE_LAST)) {
            $storage = $instance->field(DML_STORAGE_LAST) || 0;
        }

        # Get genres
        foreach my $_g ($instance->table(DML_GENRES)) {
            my $g = $_g->[0];
            next unless $g;
            push @genres, split /[\;\,]\s*/, $g;
        }

        # Get countries
        foreach my $_c ($instance->table(DML_COUNTRIES)) {
            my $c = $_c->[0];
            next unless $c;
            push @countries, split /[\;\,]\s*/, $c;
        }
    }

    # Output
    my $uri = $r->uri || ""; $uri =~ s/\/+$//;
    return $self->response_filter({
        ## General API response data
        message         => $self->error || "Ok",
        location        => $self->{location},
        uri             => $uri,
        debug           => $self->get_svar("debug") ? "on" : "off",
        year            => dtf("%YYYY"),
        #foo_attr        => $self->get_attr("foo"),

        ## Server data
        sid             => $self->get_svar("sid"),
        server_status   => $self->status,
        http_code       => $self->code,
        remote_addr     => $self->get_svar("remote_addr"),
        #name            => $name,
        #key             => $self->get_svar("key"),
        #path            => $self->get_svar("path"),
        #stamp           => $self->{stamp},
        #usr             => $usr,

        ## Database data
        database_name   => $instance_db_name,
        $db_error ? (database_error => $db_error) : (),
        $file ? (database_file => $file) : (),
        total => $cnt,
        storage_last => $storage,
        genres => [(uniq(map {$_ = ucfirst($_)} @genres))],
        countries => [(uniq(map {$_ = ucfirst($_)} @countries))],

        ## Extended data
        $self->get_svar("debug") ? (dvars => $self->{dvars}) : (),
    });
});

__PACKAGE__->register_handler( # GET /?action=error
    handler => "error",
    method  => "GET",
    path    => "/",
    query   => "action=error",
    attrs   => {
            api_method      => "error",
            content_type    => CONTENT_TYPE,
            serialize       => 1,
        },
    description => "Error handler (for tests only)",
    code    => sub {
    my $self = shift;
    my $name = shift;
    my $r = shift;
    my $q = $self->get("q");
    my $usr = $self->get("usr");

    return $self->error(Apache2::Const::SERVER_ERROR => "Test error message"); # Or 0 only!!
});

__PACKAGE__->register_handler( # GET /films
    handler => "film_list",
    method  => "GET",
    path    => "/films",
    query   => undef,
    attrs   => {
            content_type    => CONTENT_TYPE,
            serialize       => 1,
        },
    description => "Get lists of films",
    code    => sub {
    my $self = shift;
    my $name = shift;
    my $r = shift;
    my $instance = $self->instance
        or return $self->error(Apache2::Const::SERVER_ERROR => "Can't get database instance");

    # Get last from database
    my %tbl = $instance->tableh("id", DML_GET_LAST_LIST);
    if ($instance->err) {
        # $db_error = $instance->errstr
        return $self->error(Apache2::Const::SERVER_ERROR
            => sprintf("Can't get records: %s", uv2null($instance->error)));
    }
    my @last_ls = ();
    foreach my $k (sort {$b <=> $a} keys %tbl) {
        push @last_ls, $tbl{$k} if is_hash($tbl{$k});
    }

    # Output
    return $self->response_filter({
        "last"          => \@last_ls,
        "random"        => [],
    });
});

__PACKAGE__->register_handler( # GET /search
    handler => "film_search",
    method  => "GET",
    path    => "/search",
    query   => undef,
    attrs   => {
            content_type    => CONTENT_TYPE,
            serialize       => 1,
        },
    description => "Search films",
    code    => sub {
    my $self = shift;
    my $nameh = shift;
    my $r = shift;
    my $q = $self->get("q");
    my $usr = $self->get("usr");
    my $instance = $self->instance
        or return $self->error(Apache2::Const::SERVER_ERROR => "Can't get database instance");

    # Where statement
    my @where = ("1 = 1");

    # By ID
    my $id = trim($usr->{id} || 0);
    if ($id && is_int($id)) {
        push @where, sprintf("`id` = %d", $id);
    }

    # By Storage
    my $storage = trim($usr->{storage} || 0);
    if ($storage && is_int($storage)) {
        push @where, sprintf("(`storage` = %d OR `obsolete` = %d)", $storage, $storage);
    }

    # By genre
    my $genre = trim($usr->{genre} // "");
    if (length($genre)) {
        push @where, sprintf("LOWER(`genre`) LIKE LOWER('%%%s%%')", $genre);
    }

    # By country
    my $country = trim($usr->{country} // "");
    if (length($country)) {
        push @where, sprintf("LOWER(`country`) LIKE LOWER('%%%s%%')", $country);
    }

    # By date
    my $date = trim($usr->{date} || 0);
    if ($date && is_int($date)) {
        push @where, sprintf("`date` = %s", $date);
    }

    # By name
    my $name = trim($usr->{name} // "");
    if (length($name)) {
        push @where, sprintf("(LOWER(`name`) LIKE LOWER('%%%s%%') OR LOWER(`nameorig`) LIKE LOWER('%%%s%%'))", $name, $name);
    }

    # Total & Pages
    my $rows = $usr->{rows} || ROWS_PER_PAGE;
       $rows = ROWS_PER_PAGE unless is_int($rows) && $rows > ROWS_PER_PAGE_MAX;
    my $total = $instance->field(dformat(DML_SEARCH_COUNT, {WHERE => join(" AND ", @where)})) || 0;
    return $self->error(Apache2::Const::SERVER_ERROR
        => sprintf("Can't get records count: %s", uv2null($instance->error))) if $instance->err;
    my $pages = _calcPages($total, $rows);


    # Limit
    my $page = trim($usr->{page} || 1);
    my $limit = (is_int($page) && $page <= $pages)
            ? sprintf("%d, %d", ($page - 1) * $rows, $rows)
            : sprintf("%d", $rows);

    # Select
    my $sql = dformat(DML_SEARCH_FILMS, {WHERE => join(" AND ", @where), LIMIT => $limit});
    my %tbl = $instance->tableh("id", $sql);
    return $self->error(Apache2::Const::SERVER_ERROR
        => sprintf("Can't get records: %s", uv2null($instance->error))) if $instance->err;

    # Get data result
    my @ls = ();
    foreach my $k (sort {$b <=> $a} keys %tbl) {
        push @ls, $tbl{$k} if is_hash($tbl{$k});
    }

    # Output
    return $self->response_filter({
        "name"          => $name,
        "genre"         => $genre,
        "storage"       => $storage,
        "country"       => $country,
        "total"         => $total, # From database
        "page"          => $page, # Current page number
        "pages"         => $pages, # Pages number (count)
        "rows_per_page" => $rows, # RPP
        "films"         => \@ls,
    });
});

__PACKAGE__->register_handler( # POST /film
    handler => "post_film",
    method  => "POST",
    path    => "/film",
    attrs   => {
            content_type    => CONTENT_TYPE,
            serialize       => 1,
        },
    description => "Post film data",
    code    => sub {
    my $self = shift;
    my $hname = shift;
    my $r = shift;
    my $q = $self->get("q");
    my $usr = $self->get("usr");
    my $req_data = $self->get("req_data");
    #my $res_data = $self->get("res_data");
    my $instance = $self->instance
        or return $self->error(Apache2::Const::SERVER_ERROR => "Can't get database instance");

    # Validation
    my $id = tv2int(uv2zero($usr->{id}));
    my $storage = tv2int(uv2zero($usr->{storage}));
    return $self->error(Apache2::Const::HTTP_BAD_REQUEST => "Incorrect storage") unless $storage;
    my $packshot_url = $usr->{packshot_url} // "";

    # File
    my $tmpdir = File::Spec->catdir(File::Spec->tmpdir(), PREFIX);
    CTK::Util::preparedir($tmpdir)
        or return $self->error(Apache2::Const::SERVER_ERROR => "Can't prepare temp directory: $tmpdir");
    my $tmpfile = File::Spec->catfile($tmpdir, sprintf("%s.tmp", $self->get_svar("sid")));
    my $tmpposterfile = File::Spec->catfile($tmpdir, sprintf("%s-poster.tmp", $self->get_svar("sid")));

    # MIME
    my $filename = $usr->{packshot} // "";
    my $mime_sfx = "bin";
    if ($filename =~ /\.(gif|png|jpg|jpeg)$/i) {
        $mime_sfx = lc($1);
        #$mime_pfx = sprintf("data:image/%s;base64,", $mime_sfx eq 'jpg' ? "jpeg" : $mime_sfx);
    } elsif ($packshot_url =~ /\.(gif|png|jpg|jpeg)$/i) {
        $mime_sfx = lc($1);
    } elsif ($filename||$packshot_url) {
        return $self->error(Apache2::Const::HTTP_UNSUPPORTED_MEDIA_TYPE => "Incorrect mime type of packshot file: $filename");
    }

    # Uploading
    my $upload_bytes = 0;
    if ($filename) {
        if ( my $io_handle = $q->upload("packshot") ) {
            open ( MYUPLOAD, '>', $tmpfile ) or do {
                    return $self->error(Apache2::Const::SERVER_ERROR, sprintf("Can't upload packshot file %s to %s", $filename, $tmpfile));
                };
            binmode(MYUPLOAD);
            my $buffer;
            while ( my $bytesread = $io_handle->read($buffer, 60*57) ) {
                $upload_bytes += $bytesread;
                if ($upload_bytes > PACKSHOT_MAXSIZE) {
                    return $self->error(Apache2::Const::HTTP_REQUEST_ENTITY_TOO_LARGE, sprintf("The packshot file %s too large", $filename));
                }
                print MYUPLOAD $buffer; # encode_base64($buffer, "");
            }
            close MYUPLOAD;
        }
    } elsif ($packshot_url) {
        # Download by URL
        if (my $err = _wget($packshot_url, $tmpfile)) {
            return $self->error(Apache2::Const::SERVER_ERROR, sprintf("Can't download packshot %s: %s", $packshot_url, $err));
        }
        $upload_bytes = -s $tmpfile;
    }

    # Try get data from grab API
    my $name = $usr->{name} // "";
    my $grab = {};
    my $poster_sfx = "bin";
    if ($usr->{kinopoisk}) {
        $grab = _grab($self->get_dvar("apikey"), $usr->{kinopoiskid});
        return $self->error(Apache2::Const::HTTP_SERVICE_UNAVAILABLE
            => sprintf("Can't get data from kinopoisk API service: %s", $grab->{error})) if $grab->{error};
        $name = $grab->{name} if length($grab->{name});

        # Try get packshot image
        if (!$upload_bytes && $grab->{packshot}) {
            if ($grab->{packshot} =~ /\.(gif|png|jpg|jpeg)$/i) {
                $mime_sfx = lc($1);
            }
            if (my $err = _wget($grab->{packshot}, $tmpfile, $self->get_dvar("apikey"))) {
                $r->log->error(sprintf("Can't download packshot %s: %s", $grab->{packshot}, $err));
            } else {
                $upload_bytes = -s $tmpfile;
            }
        }

        # Try get poster image
        if ($grab->{poster}) {
            if ($grab->{poster} =~ /\.(gif|png|jpg|jpeg)$/i) {
                $poster_sfx = lc($1);
            }
            if (my $err = _wget($grab->{poster}, $tmpposterfile, $self->get_dvar("apikey"))) {
                $r->log->error(sprintf("Can't download poster %s: %s", $grab->{poster}, $err));
            }
        }

    } elsif ($usr->{imdb}) {
        # $grab = _grab($self->get_dvar("apikey"), $usr->{imdb});
        # No available now
    }
    return $self->error(Apache2::Const::HTTP_BAD_REQUEST => "Incorrect film title/name") unless length $name;

    # Insert empty recort and returns ID
    my $is_new = 0;
    unless ($id) {
        $instance->execute(DML_INSERT_EMPTY_RECORD, $storage, $name);
        if ($instance->err) {
            return $self->error(Apache2::Const::SERVER_ERROR
                => sprintf("Can't insert empty record: %s", uv2null($instance->error)));
        }
        $id = $instance->connect->last_insert_id("","","","");
        if ($instance->err) {
            return $self->error(Apache2::Const::SERVER_ERROR
                => sprintf("Can't get last insert id of empty record: %s", uv2null($instance->errstr)));
        }
        $is_new = 1;
    }

    # Move file to packshot dir
    my @depth = _splitDepth($id);
    my $packshot_dir = File::Spec->catdir($self->get_dvar("packshotdir"), @depth);
    my $packshot_file = File::Spec->catfile($packshot_dir, sprintf("%d.%s", $id, $mime_sfx));
    my $packshot_path = join("/", PACKSHOT_DIR, @depth, sprintf("%d.%s", $id, $mime_sfx));
    my $poster_file = File::Spec->catfile($packshot_dir, sprintf("%d-poster.%s", $id, $poster_sfx));

    # Update data
    $instance->execute(DML_SET_FILM,
        tv2int(uv2zero($usr->{obsolete})), $storage,
        $name, $grab->{nameorig} || uv2null($usr->{nameorig}),
        $grab->{genre} || $usr->{genre},
        $grab->{date} || $usr->{date},
        $grab->{imdbid} || $usr->{imdbid},
        $grab->{kinopoiskid} || $usr->{kinopoiskid},
        $grab->{country} || $usr->{country},
        $grab->{director} || $usr->{director},
        $grab->{scriptwriter} || $usr->{scriptwriter},
        $grab->{producer} || $usr->{producer},
        $grab->{operator} || $usr->{operator},
        $grab->{music} || $usr->{music},
        $grab->{artdirector} || $usr->{artdirector},
        $grab->{cut} || $usr->{cut},
        $grab->{'time'} || $usr->{'time'},
        $grab->{ratingimdb} || $usr->{ratingimdb},
        $grab->{ratingkinopoisk} || $usr->{ratingkinopoisk},
        $grab->{description} || $usr->{description},
        $id
    );
    if ($instance->err) {
        return $self->error(Apache2::Const::SERVER_ERROR
            => sprintf("Can't update record: %s", uv2null($instance->error)));
    }
    # Set packshot
    if ($upload_bytes && -f $tmpfile) {
        # Prepare directory for packshots
        CTK::Util::preparedir($packshot_dir)
            or return $self->error(Apache2::Const::SERVER_ERROR => "Can't prepare packshot directory: $packshot_dir");

        # Move temp file to packshot directory
        File::Copy::move($tmpfile, $packshot_file)
            or return $self->error(Apache2::Const::SERVER_ERROR => "Move to $packshot_file failed: $!");
        #my $slurp_data = CTK::Util::fload($tmpfile); # Slurp whole file

        # Move temp file to packshot directory
        if (-e $tmpposterfile) {
            File::Copy::move($tmpposterfile, $poster_file)
                or return $self->error(Apache2::Const::SERVER_ERROR => "Move to $poster_file failed: $!");
        }

        # Fix in database
        $instance->execute(DML_SET_PACKSHOT, $packshot_path, $id);
        if ($instance->err) {
            return $self->error(Apache2::Const::SERVER_ERROR
                => sprintf("Can't update record: %s", uv2null($instance->error)));
        }
    }

    # Output
    return $self->response_filter({
        "id"            => $id,
        "storage_last"  => $is_new ? $storage : 0,
        "upload_file"   => $tmpfile,
        "upload_bytes"  => $upload_bytes,
        "packshot_dir"  => $packshot_dir,
        "packshot_file" => $packshot_file,
        "packshot_path" => $packshot_path,
        "grab_data"     => $usr->{kinopoisk} ? "Kinopoisk" : $usr->{imdb} ? "IMDb" : "none",

    });
});

__PACKAGE__->register_handler( # DELETE /film/72
    handler => "delete_film",
    method  => "DELETE",
    path    => "/film",
    deep    => 1,
    attrs   => {
            content_type    => CONTENT_TYPE,
            serialize       => 1,
        },
    description => "Delete film by id",
    code    => sub {
    my $self = shift;
    my $hname = shift;
    my $r = shift;
    my $q = $self->get("q");
    my $usr = $self->get("usr");
    my $instance = $self->instance
        or return $self->error(Apache2::Const::SERVER_ERROR => "Can't get database instance");

    # Get ID from path or query_string
    my $path = $self->get_svar("path") // '';
    my $id = 0;
    $id = $1 if $path =~ /\/(\d{1,11})/;
    return $self->error(Apache2::Const::HTTP_BAD_REQUEST,
            sprintf("Incorrect ID. Expected is an integer value. Bad value: %s", $id)
        ) unless $id && is_int($id);

    # Delete link
    $instance->execute(DML_DELETE_FILM, $id);
    if ($instance->err) {
        return $self->error(Apache2::Const::SERVER_ERROR
            => sprintf("Can't delete record: %s", uv2null($instance->error)));
    }

    return 1; # Or 0 only!!
});

__PACKAGE__->register_handler( # GET /
    handler => "getPoster",
    method  => "GET",
    path    => "/poster",
    deep    => 1,
    description => "Get poster file",
    code    => sub {
    my $self = shift;
    my $hname = shift;
    my $r = shift;

    # Get ID from path or query_string
    my $path = $self->get_svar("path") // '';
    my $id = 0;
    $id = $1 if $path =~ /\/(\d{1,11})/;
    return $self->error(Apache2::Const::HTTP_BAD_REQUEST,
            sprintf("Incorrect ID. Expected is an integer value. Bad value: %s", $id)
        ) unless $id && is_int($id);

    my @exts = qw/gif png jpg jpeg bin/;
    my @depth = _splitDepth($id);
    my $packshot_dir = File::Spec->catdir($self->get_dvar("packshotdir"), @depth);

    # Find packshot file
    my $packshot_file;
    my $packshot_mime;
    my $packshot_size;
    foreach my $sfx (@exts) {
        my $_f = File::Spec->catfile($packshot_dir, sprintf("%d.%s", $id, $sfx));
        next unless -f $_f;
        $packshot_file = $_f;
        $packshot_mime = $sfx eq 'jpg' ? "jpeg" : $sfx;
        $packshot_size = -s $_f;
        last;
    }

    # Find poster file
    my $poster_file;
    my $poster_mime;
    my $poster_size;
    foreach my $sfx (@exts) {
        my $_f = File::Spec->catfile($packshot_dir, sprintf("%d-poster.%s", $id, $sfx));
        next unless -f $_f;
        $poster_file = $_f;
        $poster_mime = $sfx eq 'jpg' ? "jpeg" : $sfx;
        $poster_size = -s $_f;
        last;
    }

    # Not found
    unless ($poster_file || $packshot_file) {
        return $self->error(Apache2::Const::HTTP_NOT_FOUND, "No poster found");
    }

    $r->content_type(sprintf('image/%s', $poster_mime || $packshot_mime));
    my $len = $poster_mime ? $poster_size : $packshot_mime ? $packshot_size : 0;
    $r->set_content_length($len);
    my $rc = $r->sendfile($poster_file || $packshot_file, 0, $len);
    unless ($rc == APR::Const::SUCCESS) {
        return $self->error(Apache2::Const::SERVER_ERROR, "Can't send file");
    }

    return 1; # Or 0 only!!
});


#
# DB methods (internal)
#

sub _is_sqlite {
    my $self = shift;
    my $db_name = $self->get_dvar("instance_db");
    my $db = $instances{$db_name};
    return unless defined($db);
    my $driver = (DBI->parse_dsn($db->{dsn}))[1] || '';
    return ($driver =~ /sqlite/i) ? 1 : 0;
}
sub _is_mysql {
    my $self = shift;
    my $db_name = $self->get_dvar("instance_db");
    my $db = $instances{$db_name};
    return unless defined($db);
    my $driver = (DBI->parse_dsn($db->{dsn}))[1] || '';
    return ($driver =~ /mysql/i) ? 1 : 0;
}
sub _is_pg {
    my $self = shift;
    my $db_name = $self->get_dvar("instance_db");
    my $db = $instances{$db_name};
    return unless defined($db);
    my $driver = (DBI->parse_dsn($db->{dsn}))[1] || '';
    return ($driver =~ /pg/i) ? 1 : 0;
}

#
# Internal functions
#

sub _parseDBAttributes { # Example: foo=123, bar=gii,baz;qux=werwe
    my $t = shift || return ();
    my %as = ();
    for ($t ? (split /\s*[,;]\s*/, $t) : ()) {
        my ($k,$v) = (split /\s*=>?\s*|\s*\:\s*/, $_);
        $as{$k} = $v;
    }
    return %as;
}
sub _splitDepth { # Split a "qwerty" string to "q/qw/qwe"
    my $l = uv2void(shift); # String
    return () unless length $l;
    my $x = shift || XDEPTH;
    my $c = shift; # Char
    $c = "_" unless defined $c;
    $l = $c unless $l && $l =~ /^[a-zA-Z0-9_\-]+$/;
    my $pfx = $c x $x;
    if (length($l) > $x) {
        $pfx = substr($l, 0, $x);
    } else {
        $pfx = $l; #$pfx = substr($l.$pfx, 0, $x)
    }
    my @r = ();
    for (my $i = 0; $i < length($pfx); $i++) {
        push @r, substr($pfx, 0, $i+1);
    }
    return (@r);
}
sub _calcPages {
    my $i = shift || 0;
    my $l = shift || ROWS_PER_PAGE;
    return 0 unless $i;
    my $mod = $i % $l;
    my $f = ($i-$mod)/$l;
    return $mod ? $f+1 : $f;
}
sub _wget {
    my $url = shift || '';
    my $file = shift || '';
    my $apikey = shift || ''; # API key
    my $ua = LWP::UserAgent->new( # UserAgent object
        agent        => sprintf("%s/%s", __PACKAGE__, __PACKAGE__->VERSION),
        max_redirect => 10,
        requests_redirectable => ['GET','HEAD'],
        timeout      => UA_TIMEOUT,
    );
    $ua->default_header('X-API-KEY' => $apikey) if $apikey;
    my $res = $ua->mirror($url, $file);
    return $res->is_success ? undef : $res->status_line;
}
sub _grab {
    my $apikey = shift || ''; # API key
    my $kpid = shift || ''; # Kinopoisk ID
    my $imdbid = shift || ''; # IMDb ID
    my $ua = LWP::UserAgent->new( # UserAgent object
        agent        => sprintf("%s/%s", __PACKAGE__, __PACKAGE__->VERSION),
        max_redirect => 10,
        requests_redirectable => ['GET','HEAD'],
        timeout      => UA_TIMEOUT,
    );
    $ua->default_header('X-API-KEY' => $apikey) if $apikey;
    my $res = $ua->get(sprintf(API_URL_GENERAL, $kpid));
    my $result = $res->is_success ? $res->decoded_content : '';
    return {error => $res->status_line} unless $res->is_success;
    return {error => "No content transfered"} unless $result;
    Encode::_utf8_on($result);

    # Serializer
    my $sr = CTK::Serializer->new("json" , attrs => SR_ATTRS);
    my $struct = $sr->deserialize( json => $result);
    return {error => $sr->error} unless  $sr->status; # 0 - Error, 1 - Ok

    # Init new struct
    my %res = ();

    # Get countries
    my $countries = array($struct, "countries");
    my @country = ();
    foreach my $item (@$countries) {
        push @country, value($item, "country");
    }
    $res{country} = join(", ", grep {$_} @country);

    # Grt genres
    my $genres = array($struct, "genres");
    my @genre = ();
    foreach my $item (@$genres) {
        push @genre, value($item, "genre");
    }
    $res{genre} = join(", ", grep {$_} @genre);

    # Otherwise
    $res{'time'} = uv2null(value($struct, "filmLength"));
    $res{description} = uv2null(value($struct, "description"));
    $res{imdbid} = uv2null(value($struct, "imdbId"));
    $res{kinopoiskid} = uv2null(value($struct, "kinopoiskId"));
    $res{nameorig} = uv2null(value($struct, "nameOriginal"));
    $res{name} = uv2null(value($struct, "nameRu"));
    $res{poster} = uv2null(value($struct, "posterUrl"));
    $res{packshot} = uv2null(value($struct, "posterUrlPreview"));
    $res{ratingimdb} = uv2null(value($struct, "ratingImdb"));
    $res{ratingkinopoisk} = uv2null(value($struct, "ratingKinopoisk"));
    $res{date} = uv2null(value($struct, "year"));

    return {%res};
}

1;

package DBD::Sponge::db;
sub ping { 1 }

1;

__END__
